#include <stdio.h>
#include <ctype.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <getopt.h>
#include <time.h>
#include <sys/time.h>
#include <stddef.h>
#include "AwFmIndex.h"


#define RETURN_CODE_COULD_NOT_READ_KMER_FILE  -10

uint8_t aminoLookup[20]     = {'a','c','d','e','f',
                              'g','h','i','k','l',
                              'm','n','p','q','r',
                              's','t','v','w','y'};
uint8_t nucleotideLookup[4] = {'a','g','c','t'};

char indexFilenameBuffer[1024];
char sequenceFilenameBuffer[1024];
char kmerFileSrc[1024];
int numKmersToQuery;
int kmerLength;
int numThreadsInParallelQuery;
bool randSeedSetInArgs;
bool keepSuffixArrayInMemory;
bool useCountFunction;


char *globalQueryBuffer;

void parseArgs(int argc, char **argv);
void allocateQueryBuffer(struct AwFmKmerSearchList *searchList, size_t numKmers, uint8_t kmerLength);
void fillSearchListwithKmers(struct AwFmKmerSearchList *searchList, size_t numKmers, uint8_t kmerLength, struct AwFmIndex *index);
bool charIsAmbiguous(char c, enum AwFmAlphabetType alphabet);
bool getStringFromSequenceFile(FILE *openSequenceFile, size_t position, char *buffer, int kmerLength, enum AwFmAlphabetType alphabet);
void makeValidQueryFromSequenceFile(FILE *openSequenceFile, size_t sequenceLength,
uint8_t kmerLength, enum AwFmAlphabetType alphabet, char *queryBuffer);
bool readKmersFromFile(const char *kmerFileSrc, const uint8_t kmerLength, const uint64_t numKmers, struct AwFmKmerSearchList *searchList);

int main(int argc, char **argv){
//set default parameters
  strcpy(indexFilenameBuffer, "index.awfmi");
  numKmersToQuery = 10000;
  kmerLength = 12;
  randSeedSetInArgs = false;
  keepSuffixArrayInMemory = true;
  useCountFunction = false;
  kmerFileSrc[0] = 0; //null terminate the string to show that it hasn't been assigned yet.
  parseArgs(argc, argv);
//	fprintf(stderr, "args parsed\n");
  struct AwFmIndex *index;
  enum AwFmReturnCode returnCode = awFmReadIndexFromFile(&index, indexFilenameBuffer, keepSuffixArrayInMemory);
  if(returnCode < 0){
    printf("Error during index read: awFmReadIndexFromFile returned error code %i\n", returnCode);
    exit(-1);
  }
//fprintf(stderr, "index read from file\n");
  struct AwFmKmerSearchList *searchList = awFmCreateKmerSearchList(numKmersToQuery);
  if(searchList == NULL){
    printf("Error: could not allocate memory for the search data struct.\n");
    exit(-2);
  }
  allocateQueryBuffer(searchList, numKmersToQuery, kmerLength);
//fprintf(stderr, "alloc query buffers\n");
  size_t countTotal = 0;
  //printf("allocated query buffers.\n");
  double totalTime = 0.0;
  fillSearchListwithKmers(searchList, numKmersToQuery, kmerLength, index);

  //printf("beginning parallel search...");

  for(size_t i = 0; i < 4; i++){	//comment this line out (and uncomment the for loop a couple dozen lines up) to generate unique query sets for each test.
    struct timeval searchStartTime;
    gettimeofday(&searchStartTime, NULL);
    //search for the kmers
    //printf("performing one search\n");
    if(useCountFunction){
      awFmParallelSearchCount(index, searchList, numThreadsInParallelQuery);
    }
    else{
      awFmParallelSearchLocate(index, searchList, numThreadsInParallelQuery);
    }

    struct timeval searchEndTime;
    gettimeofday(&searchEndTime, NULL);
    double elapsedTime = (searchEndTime.tv_sec - searchStartTime.tv_sec);
    elapsedTime += (searchEndTime.tv_usec - searchStartTime.tv_usec) / 1000000.0;

  for(size_t queryIndex = 0; queryIndex < numKmersToQuery; queryIndex++){
    countTotal += searchList->kmerSearchData[queryIndex].count;
  }

    totalTime += elapsedTime;
    //printf("elapsed time: %f\n", elapsedTime);
  }

  totalTime /= 4;
  double averageCountPerKmer = (double)countTotal / (double)(4*numKmersToQuery);
  printf("average: query length = %u, time taken = %f seconds, avg # hits = %f\n", kmerLength, totalTime, averageCountPerKmer);

  //deallocate the query strings.
  free(globalQueryBuffer);
  awFmDeallocKmerSearchList(searchList);
  awFmDeallocIndex(index);

}


void parseArgs(int argc, char **argv){
  int option = 0;
  while((option = getopt(argc, argv, "mcdf:i:n:k:t:s:")) != -1){
    //printf("option: %c ", option);
    switch(option){
      case 'm':
        keepSuffixArrayInMemory = true;
        break;
      case 'd':
        keepSuffixArrayInMemory=false;
	break;
      case 'c':
        useCountFunction = true;
        break;
      case 'i':
        strcpy(indexFilenameBuffer, optarg);
        break;
      case 'f':
        strcpy(sequenceFilenameBuffer, optarg);
        break;
      case 'n':
        sscanf(optarg, "%i", &numKmersToQuery);
        break;
      case 'k':
        sscanf(optarg, "%i", &kmerLength);
        break;
      case 't':
        sscanf(optarg, "%i", &numThreadsInParallelQuery);
        break;
      case 's': 
        strcpy(kmerFileSrc, optarg);
	
	break;
    }
  }
}

void fillSearchListwithKmers(struct AwFmKmerSearchList *searchList, size_t numKmers, uint8_t kmerLength, struct AwFmIndex *index){
  FILE *sequenceFile = fopen(sequenceFilenameBuffer, "r");

  for(size_t i = 0; i < numKmers;i++){
    searchList->kmerSearchData[i].kmerLength = kmerLength;

    makeValidQueryFromSequenceFile(sequenceFile, index->bwtLength - 1, kmerLength,
      index->config.alphabetType, searchList->kmerSearchData[i].kmerString);
  }
}


bool charIsAmbiguous(char c, enum AwFmAlphabetType alphabet){
  if(alphabet == AwFmAlphabetNucleotide){
    for(uint8_t i = 0; i < 4; i++){
      if(tolower(c) == nucleotideLookup[i]){
        return false;
      }
    }
  }
  else{
    for(uint8_t i = 0; i < 20; i++){
      if(tolower(c) == aminoLookup[i]){
        return false;
      }
    }

  }
  return true;
}



void allocateQueryBuffer(struct AwFmKmerSearchList *searchList, size_t numKmers, uint8_t kmerLength){
  searchList->count = numKmers;
  globalQueryBuffer = malloc(numKmers * kmerLength * sizeof(char));
  for(size_t i = 0; i < numKmers; i++){
   searchList->kmerSearchData[i].kmerString = globalQueryBuffer + (i * kmerLength);
//   searchList->kmerSearchData[i].kmerString = globalQueryBuffer + i;
   searchList->kmerSearchData[i].kmerLength = kmerLength;
  }
}


void makeValidQueryFromSequenceFile(FILE *openSequenceFile, size_t sequenceLength,
  uint8_t kmerLength, enum AwFmAlphabetType alphabet, char *queryBuffer){
  size_t position = rand() % (sequenceLength - 100) + 100; //extra math to skip the first header.

  bool hasValidQuery = false;
  while(!hasValidQuery){
    hasValidQuery = getStringFromSequenceFile(openSequenceFile, position, queryBuffer, kmerLength, alphabet);
  }
//  printf("query (length %u) generated: %*.s\n", kmerLength, kmerLength, queryBuffer);
}


bool getStringFromSequenceFile(FILE *openSequenceFile, size_t position, char *buffer, int kmerLength, enum AwFmAlphabetType alphabet){
  fseek(openSequenceFile, position, SEEK_SET);
  int charactersGrabbed = 0;
  while(charactersGrabbed < kmerLength && !feof(openSequenceFile)){
    char charFromFile = fgetc(openSequenceFile);

    //if the character is at least a character...
    if(charFromFile >= 'A'&& charFromFile < 'z'){
      if(charIsAmbiguous(charFromFile, alphabet)){
        return false;
      }
      else{
        buffer[charactersGrabbed++] = charFromFile;
      }
    }
  }
//  if(feof(openSequenceFile)){
//  	printf("debug, got EOF\n");
//  }
//	printf("returning true, chars grabbed = %u, buffer = %.*s\n", charactersGrabbed, (int)charactersGrabbed, buffer);
  return true;
}


bool readKmersFromFile(const char *kmerFileSrc, const uint8_t kmerLength, const uint64_t numKmers, struct AwFmKmerSearchList *searchList){
  
  searchList->count = numKmers;
  searchList->capacity = numKmers;
  FILE *kmerFile;
  kmerFile = fopen(kmerFileSrc, "r");
  fseek(kmerFile, 0, SEEK_SET);
  if(!kmerFile){
    return false;
  }

  for(uint64_t i = 0; i < numKmers; i++){

    char *kmerBuffer = malloc(kmerLength);
    
    for(uint8_t kmerLetterIndex = 0; kmerLetterIndex < kmerLength; kmerLetterIndex++){
      if(feof(kmerFile)){
        printf("feof encountered @ i = %zu, kli=%u\n", i, (unsigned int)kmerLetterIndex);
        fclose(kmerFile);
        return false;
    }

    char c = fgetc(kmerFile);
   // printf("got char %c\n", c);
    
    kmerBuffer[kmerLetterIndex] = c;
  }
  //printf("--finished kmer buffer: %.*s\n", kmerLength, kmerBuffer);

  // the body of this while is left intentionally blank
  while((char)fgetc(kmerFile) != '\n' && !feof(kmerFile));
  
  
  searchList->kmerSearchData[i].kmerString = kmerBuffer;	
  searchList->kmerSearchData[i].kmerLength = kmerLength;
  }
  
  fclose(kmerFile);
  return true;
}
