#include <string>
#include <memory>
#include <chrono>
#include <iostream>
#include <cstdlib>

#include <seqan3/core/debug_stream.hpp>         // for debug_stream
#include <seqan3/io/sequence_file/input.hpp>    // for sequence_file_input
#include <seqan3/io/record.hpp>
#include <seqan3/io/sequence_file/output.hpp>
#include <seqan3/argument_parser/all.hpp> // for argument_parser
#include <seqan3/core/debug_stream.hpp>   // for debug_stream
#include <seqan3/search/fm_index/all.hpp>


//seemlingly needed for customizing a sdsl_wt_index_type
#include <seqan3/alphabet/nucleotide/dna4.hpp>

#include <seqan3/search/search.hpp>
#include <cereal/archives/binary.hpp>


#define DEFAULT_SA_SAMPLING_RATE 16

#ifndef SA_SAMPLING_RATE
#define SA_SAMPLING_RATE DEFAULT_SA_SAMPLING_RATE
#endif
	

//define a sdsl_wt_index_type that allows us to change the SA sampling rate, even if only by preprocessor definition.

using sdsl_wt_index_type =
  sdsl::csa_wt<sdsl::wt_blcd<sdsl::bit_vector, // Wavelet tree type
  sdsl::rank_support_v<>,
  sdsl::select_support_scan<>,
  sdsl::select_support_scan<0>>,
  SA_SAMPLING_RATE, // Sampling rate of the suffix array
  10'000'000, // Sampling rate of the inverse suffix array
  sdsl::sa_order_sa_sampling<>, // How to sample positions in the suffix array (text VS SA sampling)
  sdsl::isa_sampling<>, // How to sample positons in the inverse suffix array
  sdsl::plain_byte_alphabet>; // How to represent the alphabet



using seqan3::operator""_dna5;
using seqan3::operator""_aa27;

using std::string;

char nucAscii[5] = "agct";
char aminoAscii[21] = "acdefghiklmnpqrstvwy";
std::vector<seqan3::dna5> nucleotides = "agct"_dna5;
std::vector<seqan3::aa27> aminos = "acdefghiklmnpqrstvwy"_aa27;


int numKmers;
int kmerLength;
int dummyDebugPrintResult = 1;
bool individualKmer = false;
int maxRandLen = 8;
bool isAmino = false;
bool useCountFunction = false;
unsigned int numThreads = 1;
std::string databaseSequenceInputFileSrc = "";
long dummyReferenceValue = 0;
std::string querySaveFileSrc = "";
bool readQueryFile = false;
std::string indexFileSrc = "";
bool requestBuildIndex = false;

bool getStringFromSequenceFile(FILE *openSequenceFile, size_t position, char *buffer, int kmerLength, bool isAmino);

void makeValidQueryFromSequenceFile(FILE *openSequenceFile, size_t sequenceLength,
  uint8_t kmerLength, bool isAmino, char *queryBuffer);


bool charIsAmbiguous(char c, bool isAmino){
  if(isAmino){
    for(auto i = 0; i < 20; i++){
     if(tolower(aminoAscii[i]) == tolower(c)){
     
     	return false;
     }
    }
  }
  else{
    for(auto i = 0; i < 4; i++){
      if(tolower(nucAscii[i]) == tolower(c)){
        return false;
      }
    }
  }

  return true;
}


void makeValidQueryFromSequenceFile(FILE *openSequenceFile, size_t sequenceLength,
  uint8_t kmerLength, bool isAmino, char *queryBuffer){
  
  bool hasValidQuery = false;
  while(!hasValidQuery){
    uint32_t position = (rand() % (sequenceLength - 100)) + 100;
    hasValidQuery = getStringFromSequenceFile(openSequenceFile, position, queryBuffer, kmerLength, isAmino);
  }
}
	      
	     
bool getStringFromSequenceFile(FILE *openSequenceFile, size_t position, char *buffer, int kmerLength, bool isAmino){
  fseek(openSequenceFile, position, SEEK_SET);
  int charactersGrabbed = 0;
  while(charactersGrabbed < kmerLength && !feof(openSequenceFile)){
    char charFromFile = fgetc(openSequenceFile);

    //if the character is at least a character...
    if(charFromFile >= 'A'&& charFromFile < 'z'){
    
      if(charIsAmbiguous(charFromFile, isAmino)){
        return false;
      }
      else{
        buffer[charactersGrabbed++] = charFromFile;
      }
    }
  }
  if(feof(openSequenceFile)){
    seqan3::debug_stream << "debug, got EOF\n";
    return false;
  }
    
  return true;
}
                                                                         

std::vector<std::vector<seqan3::dna5>> generateNucQueries(const int numKmers, const int kmerLength, char *fastaSrc){
  //seqan3::debug_stream << " generating " << numKmers << " dna queries of length "<< kmerLength <<"\n";
 

  FILE *fasta = fopen(fastaSrc, "r");
  if(!fasta){
   seqan3::debug_stream << "ERROR: failure to open sequence file '"<< fastaSrc<< "'.\n";
  }
  auto queries =  std::vector<std::vector<seqan3::dna5>>();
  queries.resize(numKmers);
  char kmerBuffer[kmerLength+1];
  int i = 0;
  while(i < numKmers){
    size_t position = (rand() % 990000000) + 100;
    bool querySuccess = getStringFromSequenceFile(fasta, position, kmerBuffer, kmerLength, false); 
    if(querySuccess){  
      
      queries[i].resize(kmerLength);
      for(auto j = 0; j < kmerLength; j++){
        queries[i][j].assign_char(kmerBuffer[j]);
      }
      	
      i++;
    }
  }

  //seqan3::debug_stream << "finished generating queries\n";
  fclose(fasta);

  return queries;
}


std::vector<std::vector<seqan3::aa27>> generateAminoQueries(const int numKmers, const int kmerLength, char *fastaSrc){
  auto queries = std::vector<std::vector<seqan3::aa27>>();
  queries.resize(numKmers);  

  FILE *fasta = fopen(fastaSrc, "r");
  if(!fasta){
    seqan3::debug_stream << "ERROR: failure to open sequence file '"<< fastaSrc<< "'.\n";
  }

  char kmerBuffer[kmerLength+1];
  int i = 0;
  while( i < numKmers){
    size_t position = (rand() % 190000000)+200;
    bool querySuccess = getStringFromSequenceFile(fasta, position, kmerBuffer, kmerLength, true);
    if(querySuccess){
      queries[i].resize(kmerLength);

      for(auto j = 0; j < kmerLength; j++){
        queries[i][j].assign_char(kmerBuffer[j]);
      }  
      i++;
    }
  }

  fclose(fasta);
  
  
  return queries;
}


void saveNucQueriesToFile(std::vector<std::vector<seqan3::dna5>> &dnaQueries, const std::string &fileSrc){
  std::fstream fs;
  fs.open(fileSrc, std::fstream::out | std::fstream::trunc);
  for(auto query: dnaQueries){
    for(auto c: query){
      fs << c.to_char();
    }
    fs << "\n";
  }
}


void readNucQueriesFromFile(std::vector<std::vector<seqan3::dna5>> &dnaQueries, long numQueries, int kmerLength, const std::string &fileSrc){
  
  std::ifstream stream(fileSrc);
  dnaQueries.resize(numQueries);

  for(long i = 0; i < numQueries; i++){
    std::string query;
    std::getline(stream, query);
    dnaQueries[i].resize(kmerLength);

    for(int j = 0; j < kmerLength; j++){
      dnaQueries[i][j].assign_char(query[j]);
    }
  }
}


int main (int argc, char ** argv)
{
  //set seqan to use cout as the output stream.
  seqan3::debug_stream.set_underlying_stream(std::cout);


  seqan3::debug_stream << "sa sampling rate is " << std::to_string(SA_SAMPLING_RATE) << "\n";
  std::srand(std::time(0));
  
  //seqan3::debug_stream << "application begin\n";
  seqan3::argument_parser parser("seqanIndex", argc, argv);

  parser.add_option(numKmers,'n', "num_kmers", "number of kmers to query.");
  parser.add_option(kmerLength, 'k', "kmer_length", "length of kmer to query.");
  parser.add_flag(isAmino, 'a', "amino", "which alphabet to generate and expect. 0 for nucleotide, 1 for amino.");
  parser.add_option(databaseSequenceInputFileSrc, 'f', "fasta_input","location of the file to use to build the index, and to use to generate queries.");
  parser.add_option(dummyDebugPrintResult, 'd', "dummy", "dummy debug print result value (set to 0 to print");
  parser.add_flag(useCountFunction, 'c', "count", "0 to perform the locate function, 1 to perform a count function.");
  parser.add_option(numThreads,'t', "num_threads", "number of threads to run search with");
  parser.add_option(querySaveFileSrc, 's', "query_save_file_src", "file src to save the generated queries to. setting this option forces this to save the queries instead of searching for them.");
  parser.add_flag(readQueryFile, 'r', "read_query_file","read the queries from the file src given in querySaveFileSrc. This flag is illegal unless the -s option is set."); 
  parser.add_option(indexFileSrc, 'i', "index_file_src", "file location of the index to read for this test.");
  parser.add_flag(requestBuildIndex, 'b', "build_index", "tells the program to generate an index file from the given fasta file (-f) and save it to the given index file src (-i-).");
  
  try{
    parser.parse();
  }
  catch (seqan3::argument_parser_error const & ext)
  {
    seqan3::debug_stream << "[PARSER ERROR] " << ext.what() << '\n';
    return 0;
  }


  if(indexFileSrc.empty()){
    seqan3::debug_stream << "no index file specified, one must be specified using the -i flag.";
    exit(EXIT_FAILURE);
   }
   if(databaseSequenceInputFileSrc.empty()){
     seqan3::debug_stream << "no fasta file specified, one must be specified using the -f flag.";
    exit(EXIT_FAILURE);
  }

  //build the index if requested
  if(requestBuildIndex){
    if(isAmino){
    //the amino sequence file needs the <sequence_file_input_default_traits_aa> trait in oder to read as protein data.
      seqan3::sequence_file_input<seqan3::sequence_file_input_default_traits_aa> fastaContents{databaseSequenceInputFileSrc};
           //seqan3 only has examples of using a for-each loop to get sequences out of the file, so I guess that's what we gotta do.
     
       for(auto & record: fastaContents){
      //we're only going to grab the first entry, so we'll quit out after this.
        seqan3::fm_index<seqan3::aa27, seqan3::text_layout::single, sdsl_wt_index_type> index(seqan3::get<seqan3::field::seq>(record));
        //seqan3::fm_index<seqan3::aa27, seqan3::text_layout::single> index(record.id());
	std::ofstream stream {indexFileSrc, std::ios::binary};
        cereal::BinaryOutputArchive oarchive{stream};
        oarchive(index);
	break;
      }
    }
    else{
      seqan3::sequence_file_input fastaContents{databaseSequenceInputFileSrc};
    //seqan3 only has examples of using a for-each loop to get sequences out of the file, so I guess that's what we gotta do.
      for(auto & record: fastaContents){
        //we're only going to grab the first entry, so we'll quit out after this.
        seqan3::fm_index<seqan3::dna5, seqan3::text_layout::single, sdsl_wt_index_type> index(seqan3::get<seqan3::field::seq>(record));
        //seqan3::fm_index<seqan3::dna5, seqan3::text_layout::single> index(record.id());
	std::ofstream stream {indexFileSrc, std::ios::binary};
        cereal::BinaryOutputArchive oarchive{stream};
        oarchive(index);
        break;
      }
    }
    seqan3::debug_stream << "index creation finished, index file is located at " << indexFileSrc << ".\n";
    exit(EXIT_SUCCESS);
  } 


  auto const noErrorsConfig = seqan3::search_cfg::default_configuration; 

  seqan3::fm_index<seqan3::dna5, seqan3::text_layout::single, sdsl_wt_index_type> nucIndex;
  seqan3::fm_index<seqan3::aa27, seqan3::text_layout::single, sdsl_wt_index_type> aminoIndex;

  if(isAmino){
    std::ifstream is{indexFileSrc, std::ios::binary};
    cereal::BinaryInputArchive iarchive{is};
    iarchive(aminoIndex);
  }
  else{
    std::ifstream is{indexFileSrc, std::ios::binary};
    cereal::BinaryInputArchive iarchive{is};
    iarchive(nucIndex);
  }

  struct timeval searchStartTime, searchEndTime;
  double elapsedTime = 0.0;

  std::vector<std::vector<seqan3::dna5>> nucQueries;
  std::vector<std::vector<seqan3::aa27>> aminoQueries;
  char sequenceFileAsCstr[1024];
  strcpy(sequenceFileAsCstr, databaseSequenceInputFileSrc.c_str());
  if(isAmino)    aminoQueries = generateAminoQueries(numKmers, kmerLength, sequenceFileAsCstr);	     
  else           nucQueries = generateNucQueries(numKmers, kmerLength, sequenceFileAsCstr);
  
  size_t countTotal = 0;
  for(auto i = 0; i < 4; i++){  //this line is here in case you want to only generate the queries once per benchmark
    gettimeofday(&searchStartTime, NULL);
    //auto chronoStartTime = std::chrono::system_clock::now();
    if(useCountFunction){
      //count
      if(isAmino){
        for(auto && query: aminoQueries){
          auto cursor = aminoIndex.cursor();
          cursor.extend_right(query);
	  auto count = cursor.count();
	  countTotal+= count;
	}
      }
      else{
        for(auto && query:nucQueries){
	  auto cursor = nucIndex.cursor();
	  cursor.extend_right(query);
	  auto count = cursor.count();
	  countTotal += count;
  	}
      }

    }
    else{
    //locate
      if(isAmino){
      	//seqan3::debug_stream << "performing amino locate\n";
        auto searchResult = seqan3::search(aminoQueries, aminoIndex);//, noErrorsConfig);//, searchConfig);

        for(auto result: searchResult){
	  countTotal++;
	    dummyReferenceValue += result.reference_begin_position();
        }
      }
      else{
        auto searchResult = seqan3::search(nucQueries, nucIndex);//, noErrorsConfig);//, searchConfig);

        for(auto result: searchResult){
	  countTotal++;
	  dummyReferenceValue += result.reference_begin_position();
        } 
      }
    }
    //std::chrono::duration<double> chronoDuration(std::chrono::system_clock::now() - chronoStartTime);
    gettimeofday(&searchEndTime, NULL); 
    double testTimeTotal = searchEndTime.tv_sec - searchStartTime.tv_sec;
    testTimeTotal += (searchEndTime.tv_usec - searchStartTime.tv_usec) / 1000000.0;
    elapsedTime += searchEndTime.tv_sec - searchStartTime.tv_sec;
    elapsedTime += (searchEndTime.tv_usec - searchStartTime.tv_usec) / 1000000.0; 
    //seqan3::debug_stream << ", count: "<< countTotal <<", test time "<< testTimeTotal<<".\n";
  }

  elapsedTime /= 4;
  double countTotalAsDouble = countTotal;
  countTotalAsDouble /=  4 * numKmers;
  if(dummyReferenceValue == 999){
    seqan3::debug_stream << "debug message, please ignore this line.\n";
  }
  seqan3::debug_stream << "average time: " << elapsedTime<< " average # hits/query: "<< countTotalAsDouble << "\n";
  return 0;
}


