#include <string>
#include <memory>
#include <chrono>
#include <iostream>
#include <cstdlib>

#include <seqan3/core/debug_stream.hpp>         // for debug_stream
#include <seqan3/io/sequence_file/input.hpp>    // for sequence_file_input
#include <seqan3/io/record.hpp>
#include <seqan3/io/sequence_file/output.hpp>
#include <seqan3/argument_parser/all.hpp> // for argument_parser
#include <seqan3/core/debug_stream.hpp>   // for debug_stream
#include <seqan3/search/fm_index/all.hpp>

#include <seqan3/search/search.hpp>
#include <cereal/archives/binary.hpp>


#define DEFAULT_SA_SAMPLING_RATE 16

#ifndef SA_SAMPLING_RATE
#define SA_SAMPLING_RATE DEFAULT_SA_SAMPLING_RATE
#endif


//define a sdsl_wt_index_type that allows us to change the SA sampling rate, even if only by preprocessor definition.

using sdsl_wt_index_type =
  sdsl::csa_wt<sdsl::wt_blcd<sdsl::bit_vector, // Wavelet tree type
  sdsl::rank_support_v<>,
  sdsl::select_support_scan<>,
  sdsl::select_support_scan<0>>,
  4, // Sampling rate of the suffix array
  10'000'000, // Sampling rate of the inverse suffix array
  sdsl::sa_order_sa_sampling<>, // How to sample positions in the suffix array (text VS SA sampling)
  sdsl::isa_sampling<>, // How to sample positons in the inverse suffix array
  sdsl::plain_byte_alphabet>; // How to represent the alphabet



using seqan3::operator""_dna5;
using seqan3::operator""_aa27;

using std::string;

std::vector<seqan3::dna5> nucQuery = "TATCTCTGACACGT"_dna5;
std::vector<seqan3::aa27> aminoQuery = "GAFFPQ"_aa27;

const int numKmers = 1;
int dummyDebugPrintResult = 1;
bool useCountFunction = false;
bool isAmino = false;
unsigned int numThreads = 1;
long dummyReferenceValue = 0;
std::string indexFileSrc;

int main (int argc, char ** argv)
{
  //set seqan to use cout as the output stream.
  seqan3::debug_stream.set_underlying_stream(std::cout);
  std::srand(std::time(0));
  
  //seqan3::debug_stream << "application begin\n";
  seqan3::argument_parser parser("seqanIndex", argc, argv);

  parser.add_option(dummyDebugPrintResult, 'd', "dummy", "dummy debug print result value (set to 0 to print");
  parser.add_flag(useCountFunction, 'c', "count", "0 to perform the locate function, 1 to perform a count function.");
  parser.add_option(indexFileSrc, 'i', "index_file_src", "file location of the index to read for this test.");
  parser.add_flag(isAmino, 'a', "amino", "setting the -a flag uses the amino query instead of the dna query.");
 
  try{
    parser.parse();
  }
  catch (seqan3::argument_parser_error const & ext)
  {
    seqan3::debug_stream << "[PARSER ERROR] " << ext.what() << '\n';
    return 0;
  }


  if(indexFileSrc.empty()){
    seqan3::debug_stream << "no index file specified, one must be specified using the -i flag.";
    exit(EXIT_FAILURE);
   }


  auto const noErrorsConfig = seqan3::search_cfg::default_configuration; 

  seqan3::fm_index<seqan3::dna5, seqan3::text_layout::single, sdsl_wt_index_type> nucIndex;
  seqan3::fm_index<seqan3::aa27, seqan3::text_layout::single, sdsl_wt_index_type> aminoIndex;
  std::ifstream is{indexFileSrc, std::ios::binary};
  cereal::BinaryInputArchive iarchive{is};
  if(isAmino){
    iarchive(aminoIndex);
  }
  else{
    iarchive(nucIndex);
  }


  //std::vector<std::vector<seqan3::dna5>> nucQueries;
  
  size_t countTotal = 0;
  for(auto i = 0; i < 4; i++){  //this line is here in case you want to only generate the queries once per benchmark
    if(useCountFunction){
      //count
      if(isAmino){
        auto cursor = aminoIndex.cursor();
        cursor.extend_right(aminoQuery);
        auto count = cursor.count();
        countTotal += count;
      }
      else{
        auto cursor = nucIndex.cursor();
        cursor.extend_right(nucQuery);
        auto count = cursor.count();
        countTotal += count;
      }
    }
    else{
    //locate
      if(isAmino){
        auto searchResult = seqan3::search(aminoQuery, aminoIndex);
        for(auto result: searchResult){
          countTotal++;
        }
      }
      else{
        auto searchResult = seqan3::search(nucQuery, nucIndex);//, noErrorsConfig);//, searchConfig);
        for(auto result: searchResult){
  	  countTotal++;
        } 
      }
    }
  }

  double countTotalAsDouble = (double)countTotal / (double)4;
  countTotalAsDouble /=  numKmers;
  seqan3::debug_stream << " average # hits/query: "<< countTotalAsDouble << "\n";
  return 0;
}


